% Copyright 2007 Noah B. Jacobsen.
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Lesser General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
% Lesser General Public License for more details.
% 
% You should have received a copy of the GNU Lesser General Public
% License along with this program.  If not, see <http://www.gnu.org/licenses/>.
%
function [dv,pv,dc,pc,cs2] = match_var(R,dv,p_eg,d_eg,R0,dv0,pv0,dc0,pc0,xx,tol,EPS,filename,p1_n,cs2)

dc_max = 25;
n0 = sum(pv0./dv0);

I_ch = R + tol;
I_v = get_var_fn(dv,I_ch,xx);

npr = (1-R/R0)/(1-R); %% new parity ratio
if(p_eg>npr)
  error('\n Too many PEG parity rows.');
end

pc0_n = pc0./dc0/sum(pc0./dc0); 
pc0_n = pc0_n*(1-npr);

pc_n = zeros(1,dc_max);

%%%Compute split parity distribution
u = npr-p_eg;
while( u > EPS )

  for d = length(dc0):-1:1
    pr = pc0_n(d);
  
    if( u > EPS )
      if( u < pr )

        pc_n(dc0(d)) = pc_n(dc0(d)) + (pr - u); 
        pc_n(floor(dc0(d)/2)+1) = pc_n(floor(dc0(d)/2)+1) + u;
        pc_n(ceil(dc0(d)/2)+1) = pc_n(ceil(dc0(d)/2)+1) + u;
        u = -1;
      end
      
      if( u >= pr )
        pc_n(floor(dc0(d)/2)+1) = pc_n(floor(dc0(d)/2)+1) + pr;
        pc_n(ceil(dc0(d)/2)+1) = pc_n(ceil(dc0(d)/2)+1) + pr;
        u = u - pr;
      end
  
    else
      pc_n( dc0(d)) = pc_n( dc0(d)) + pr;
    end
  end

  %%If need to split multiple times
  if( u>EPS )
    pc0_n = [];
    dc0 = [];
    
    index = 1;
    for d = 1:dc_max
      if ( pc_n(d) > 0 )

        pc0_n( index) = pc_n(d);
        dc0( index) = d;
        index = index+1;
      end
    end    
    
    pc_n = zeros(1,dc_max);
  end
end

%%Add edge growth parities
p = (ceil(d_eg)-d_eg);
pc_n(ceil(d_eg)) = pc_n(ceil(d_eg)) + (1-p)*p_eg;
pc_n(floor(d_eg)) = pc_n(floor(d_eg)) + p*p_eg;

%%Compute edge distribution
index = 1;
for d=1:dc_max
  if ( pc_n(d) > 0 )

    pc(index) = pc_n(d)*d;
    dc(index) = d;
    index = index+1;
  end
end
pc = pc/sum(pc);
n1 = sum(pc./dc)/(1-R);

%%Target check function
I_c = get_chk_fn(dc,xx);
I_c = pc*I_c;

Nv = length(dv);
pv_n = zeros(1,Nv);
  
for d = 1:length(dv0)
  for n = 1:Nv

    if( dv0(d)==dv(n))
      pv_n(n) = pv0(d)/dv0(d)/n0;
    end
  end
end

%% rate-compatible constraint
P0 = (1-R/R0)+(R/R0)*cumsum(pv_n); 
%% split parity constraint
cs2 = cs2*(R/R0) + (1-R)*(npr-p_eg);

% $$$ R = 1-((1./dc)*pc')/((1./dv)*pv');
%Ceq = [ones(1,Nv); 1./dv; [1/n1 zeros(1,Nv-1)]];
%deq = [1; (1./dc)*pc'/(1-R); p1_n];
Ceq = [ones(1,Nv); 1./dv];
deq = [1; (1./dc)*pc'/(1-R)];
       
lb = [zeros(Nv,1)];
ub = [ones(Nv,1)];

C = I_v';
d = I_c';
A = [-I_v'; %% convergence constraint
     tril(ones(Nv,1)*(1./dv)/n1); %% rc constraint
     %-[0 1/2/n1 zeros(1,Nv-2)]; %% #2s > cs2
     %[0 1/2/n1 zeros(1,Nv-2)];  %% #2s < M
    ];
b = [-I_c';P0';%-cs2;(1-R)*n1
    ];

options = optimset('lsqlin');
options = optimset(options,'MaxIter',500);
options = optimset(options,'Display','Iter');

pv=lsqlin(C,d,A,b,Ceq,deq,lb,ub,[],options);
%pv = lsqlin(C,d,A,b,Ceq,deq,lb,ub);
plot(xx,I_c,'b-',xx,pv'*I_v,'g--')

pv_n = pv'./dv/sum(pv'./dv);
dv = dv(find(pv_n>EPS));
pv_n = pv_n(find(pv_n>EPS));

pv_n = pv_n/sum(pv_n);
pv = pv_n.*dv/sum(pv_n.*dv);
testrc;

if(p1_n>0)
  p1 = [p1_n pv_n(1)-p1_n pv_n(2:end)];
  d1 = [1 dv];
else
  p1 = pv_n;
  d1 = dv;
end

outfile = fopen(filename,'a');
fprintf(outfile,'%1.7e %1.7e\n%d\n',R,npr-p_eg,length(d1));
for n=1:length(d1),fprintf(outfile,'%d ',d1(n));end;
fprintf(outfile,'\n');
for n=1:length(d1),fprintf(outfile,'%1.7e ',p1(n));end;
fprintf(outfile,'\n\n');
fclose(outfile);
