/* Copyright 2007 Noah B. Jacobsen.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/******************************************************************************/
/* *******Includes code from:*********                                        */
/* Copyright (c) 2000, 2001 by Radford M. Neal                                */
/*                                                                            */
/* Permission is granted for anyone to copy, use, modify, or distribute this  */
/* program and accompanying programs and documents for any purpose, provided  */
/* this copyright notice is retained and prominently displayed, along with    */
/* a note saying that the original programs are available from Radford Neal's */
/* web page, and note is made of any changes made to the programs.            */
/******************************************************************************/
/* ENC.C - Encoding procedures. */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "rand.h"
#include "alloc.h"
#include "mod2sparse.h"
#include "mod2dense.h"
#include "mod2convert.h"
#include "ncode.h"
#include "enc.h"


/* The procedures in this module obtain the generator matrix to use for
   encoding from the global variables declared in ncode.h */

void harq_encode
( char *sblk,
  char *cblk,
  mod2dense *u,
  mod2dense *v,
  mod2dense **codeWord,
  mod2dense **harqBits
)
{
  int i,j,k,l,K=N-M;
  int bit;
  mod2dense *temp_word;

  temp_word = mod2dense_allocate(N,1);

  /* Copy source bits to the systematic part of the coded block. */

  for (j = codeLength[0]-K; j<codeLength[0]; j++)
  { cblk[cols[j]] = sblk[j-(codeLength[0]-K)];
  }

  /* Multiply by Inv(A) X B to produce check bits. */

  for (j = 0; j<K; j++)
  { mod2dense_set(u,j,0,sblk[j]);
  }
  
  mod2dense_multiply(G[0],u,v);

  /* Copy check bits to the right places in the coded block. */

  for (j = 0; j<codeLength[0]-K; j++)
  { cblk[cols[j]] = mod2dense_get(v,j,0);
  }

  /* Copy initial codeword with right ordering. */

  for (k=0; k<codeLength[0]; k++){
    bit = (int)cblk[codeLength[0]-1-k];

    for(j=0; j<=harqNum; j++)
      mod2dense_set(codeWord[j],k,0,bit);
  }

  /* Compute H-ARQ code-bits */

  for (l=1; l<=harqNum; l++){
    mod2dense_multiply(G[l],codeWord[l-1],harqBits[l-1]);

    for (k=codeLength[l-1]; k<codeLength[l]; k++){
      bit = mod2dense_get(harqBits[l-1],k-codeLength[l-1],0);

      for(j=l; j<=harqNum; j++)
	mod2dense_set(codeWord[j],k,0,bit);
    } 
  }
  
  /* Copy final codeword  */
  for (j=0; j<codeLength[harqNum]; j++)
    cblk[j] = mod2dense_get(codeWord[harqNum],j,0);

}
