/* Copyright 2007 Noah B. Jacobsen.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/* Import the parity check matrix to matlab */

#include "mex.h"

#include "alloc.h"
#include "mod2sparse.h"
#include "mod2dense.h"
#include "mod2convert.h"
#include "ncode.h"

void usage(void);

void
mexFunction( int nlhs, mxArray *plhs[],
             int nrhs, const mxArray *prhs[] )
{
  char *pchk_file;
  double *pchk_output, *pr;
  int n, *ir, *jc, k = 0, num_ones = 0, is_complex = 0;
/*   mxLogical *pl; */
  mod2entry *e;

  if( nrhs != 1 || nlhs!=1 || !mxIsChar(prhs[0]) ) usage();

  pchk_file = mxArrayToString( prhs[0] );
  read_pchk( pchk_file );
  
  for( n = 0; n<N; n++ ){
    e = mod2sparse_first_in_col(H,n);
    while( !mod2sparse_at_end(e) ){
      num_ones += 1;
      e = mod2sparse_next_in_col(e);
    }
  }  
/*   mexPrintf("the number of ones is %d\n", num_ones); */
  
/*   plhs[0] = mxCreateSparseLogicalMatrix(M,N,num_ones); */
/*   pl = mxGetLogicals(plhs[0]); */
  plhs[0] = mxCreateSparse(M,N,num_ones,is_complex);
  pr = mxGetPr(plhs[0]);
  ir = mxGetIr(plhs[0]);
  jc = mxGetJc(plhs[0]);
 
  for( n = 0; n<N; n++ ){
    e = mod2sparse_first_in_col(H,n);
    jc[n] = k;
    while( !mod2sparse_at_end(e) ){
      pr[k] = 1;
      ir[k] = mod2sparse_row(e);
      e = mod2sparse_next_in_col(e);     
      k += 1;
    }
  }
  jc[N] = k;

  return;
}

void usage(void){ 
  mexPrintf("\nUsage:\n\n");
  mexPrintf("[output_pchk_matrix] = import_pchk('pchk_file'\n");
  mexErrMsgTxt("Please try again");
}
