#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "alloc.h"
#include "mod2sparse.h"
#include "mod2dense.h"
#include "mod2convert.h"
#include "rcode.h"
#include "check.h"
#include "dec.h"
#include "enc.h"
#include "mex.h"

#define Soft 100
#define sign(x) pow(-1,x<0)
#define abs(x) x*sign(x)

void usage(void);
void decode(double *,double *,int,double *,double *,double *,double *);

void decode
( double *a, 
  double *b, 
  int blocks, 
  double *c_hat, 
  double *u_hat, 
  double *block_errs, 
  double *info_map 
)
{ char *dblk, *pchk;
  double *lratio;
  double *bitpr;
  
  unsigned iters;	
  double tot_iter;	
  double chngd, tot_changed;

  int tot_valid, block_no;
  int valid, invalid;

  int i, j, k;

  dblk   = chk_alloc (N, sizeof *dblk);
  lratio = chk_alloc (N, sizeof *lratio);
  pchk   = chk_alloc (M, sizeof *pchk);
  bitpr  = chk_alloc (N, sizeof *bitpr);

  tot_iter = 0;
  tot_valid = 0;
  tot_changed = 0;

  for (block_no = 0; block_no<blocks; block_no++){
    
    for(i = 0; i<N; i++){
 
      lratio[i] = exp(a[block_no*N+i]);
    }

    /* Call decoder. */
    iters = prprp_decode (H, lratio, dblk, pchk, bitpr);
    
    /*     mexPrintf("%d iterations were necessary\n",iters); */
  
    /* See if it worked, and how many bits were changed */
    
    invalid = check(H,dblk,pchk);
    valid = invalid==0;
    chngd = changed(lratio,dblk,N);
    
    //mexPrintf("%d/%d check errors\n",invalid,M); 
    
    tot_iter += iters;
    tot_valid += valid;
    tot_changed += chngd;

    /* Write bit probabilities */

    for (j = 0; j<N; j++){
      b[N*block_no+j] = -log(bitpr[j]) - a[N*block_no+j];
      
     /*  if(j<10) */
/* 	mexPrintf("\n%f = %f*%f ",b[N*block_no+j],sign(b[N*block_no+j]),abs(b[N*block_no+j])); */

      if( abs(b[N*block_no+j]) > Soft )
	b[N*block_no+j] = Soft*sign(b[N*block_no+j]);
    }

    /* Write decoded code word if called for */

    if (c_hat)
      for (j = 0; j<N; j++ )
	c_hat[N*block_no+j] = bitpr[j]<1;

    /* Write decoded info bits if called for */

    if (u_hat)
      for (j = 0; j<N-M; j++ )
	u_hat[(N-M)*block_no+j] = bitpr[(int)info_map[j]-1]<1;
  }

  if (block_errs) block_errs[0] = blocks-tot_valid;

/*   fprintf(stderr,"Decoded %d blocks, %d valid.  Average %.1f iterations, %.0f%% bit changes\n", */
/* 	  block_no, tot_valid, (double)tot_iter/block_no, */
/* 	  100.0*(double)tot_changed/(N*block_no)); */

  mxFree(dblk);
  mxFree(lratio);
  mxFree(pchk);
  mxFree(bitpr);

}


void
mexFunction( int nlhs, mxArray *plhs[],
             int nrhs, const mxArray *prhs[] )
{
  char *pchk_file, junk; 
  double *a, *b, *c_hat, *u_hat, *block_errs, *info_map;
  int info, coded, blocks; 
  
  fflush(stdout);
  if( nrhs<3 || nrhs>4 )
    usage();

  if ( !mxIsChar( prhs[0] ) ) usage();
  pchk_file = mxArrayToString( prhs[0] );
  read_pchk( pchk_file );

  if ( !mxIsDouble( prhs[1] ) ) usage();
  max_iter = (int)mxGetScalar( prhs[1] );

  if ( !mxIsDouble( prhs[2] ) ) usage();
  if( mxGetM(prhs[2]) != 1 )
    mexErrMsgTxt("Please use row vector for input");

  coded = mxGetN( prhs[2] );
  blocks = (int)coded/N;

  if( (double)blocks != (double)coded/N )
    mexErrMsgTxt("Not an integer number of blocks");
  
  info = blocks*(N-M);
  a = mxGetPr( prhs[2] );

  if ( nlhs!=4 && nrhs==4 ) usage();
  if ( nrhs==4 )
    info_map = mxGetPr( prhs[3] );
 
  
  /* Allocating output */

  if ( nlhs==0 || nlhs>4 ) usage();
  
  plhs[0] = mxCreateDoubleMatrix(1,coded,mxREAL);
  b = mxGetPr( plhs[0] );

  if ( nlhs>1 ){
    plhs[1] = mxCreateDoubleMatrix(1,blocks,mxREAL);
    block_errs = mxGetPr( plhs[1] );
  }
  else{
    block_errs = 0;
    c_hat = 0;
    u_hat = 0;
  }

  if ( nlhs>2 ){
    plhs[2] = mxCreateDoubleMatrix(1,coded,mxREAL);
    c_hat = mxGetPr( plhs[2] );
  }
  else{
    c_hat = 0;
    u_hat = 0;
  }

  if ( nlhs == 4  ){
    plhs[3] = mxCreateDoubleMatrix(1,info,mxREAL);
    u_hat = mxGetPr( plhs[3] );

    if ( nrhs==3){
      info_map   = chk_alloc (N-M, sizeof *info_map);
      
      int jj;
      for(jj=0;jj<N-M;jj++)
	info_map[jj] = jj + M;
    }
    
  }
  else u_hat = 0;


  decode(a,b,blocks,c_hat,u_hat,block_errs,info_map);
  
  mxFree(pchk_file);
  return;
}

void usage(void){ 
  mexPrintf("\nUsage:\n\n");
  mexPrintf("[ info_app [, block_errs] [, code_words] [, decoded_info] ]\n");
  mexPrintf("   = decode ( 'pchk_file', max_iterations, channel_output [, message_bit_index] )\n\n");
}



