#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "alloc.h"
#include "mod2sparse.h"
#include "mod2dense.h"
#include "mod2convert.h"
#include "rcode.h"
#include "enc.h"
#include "mex.h"

void usage(void);
void encode(double *,double *,int,int);

void
encode( double *info_bits, double *c, int info, int blocks )
{
  mod2dense *u, *v;
  char *sblk, *cblk, *chks;
  int i, k, n, t;
  
  /* Allocate needed space. */

  if (type=='d')
  { u = mod2dense_allocate(N-M,1);
    v = mod2dense_allocate(M,1);
  }

  if (type=='m')
  { u = mod2dense_allocate(M,1);
    v = mod2dense_allocate(M,1);
  }

  sblk = chk_alloc (N-M, sizeof *sblk);
  cblk = chk_alloc (N, sizeof *cblk);
  chks = chk_alloc (M, sizeof *chks);

 
  /* Encode successive blocks */
  
  for (n = 0; n<blocks; n++){
    
    /* Read block from source var */
    
    for (t = 0; t<N-M; t++)
      sblk[t] = (char) info_bits[n*(N-M)+t];



    /* Compute encoded block. */

    switch (type)
    { case 's':
      { sparse_encode (sblk, cblk);
        break;
      }
      case 'd':
      { dense_encode (sblk, cblk, u, v);
        break;
      }
      case 'm':
      { mixed_encode (sblk, cblk, u, v);
        break;
      }
    }

    /* Check that encoded block is a code word. */

    mod2sparse_mulvec (H, cblk, chks);

    for (i = 0; i<M; i++) 
    { if (chks[i]==1)
      { fprintf(stderr,"Output block %d is not a code word!  (Fails check %d)\n",n,i);
        abort(); 
      }
    }

    /* Write encoded block to encoded output var */
    
    for (t = 0; t<N; t++)
      c[n*N+t] = (double)cblk[t];

  }
}

void
mexFunction( int nlhs, mxArray *plhs[],
             int nrhs, const mxArray *prhs[] )
{
  char *pchk_file, *gen_file; 
  double *u, *c, *info_map;
  int info, coded, blocks, k;
  
  if( nrhs != 3  || nlhs==0 || nlhs >2 ) usage();
  if( !mxIsChar(prhs[0]) || !mxIsChar(prhs[1]) )
    mexErrMsgTxt("File names must be given as strings");
  if( mxGetM(prhs[2]) != 1 )
    mexErrMsgTxt("Please use row vector for input");
  
  pchk_file = mxArrayToString( prhs[0] );
  gen_file = mxArrayToString( prhs[1] );
  
  read_pchk(pchk_file);
  read_gen(gen_file,0,0);

  
  
  if( N<=M ) 
    mexErrMsgTxt("Can't encode if number of bits is not greater than number of checks");
  
  info = mxGetN( prhs[2] );
  blocks = (int)info/(N-M);


  if( (double)blocks != (double)info/(double)(N-M) )
    mexErrMsgTxt("Not an integer number of blocks");
  
  coded = blocks*N;
  u = mxGetPr( prhs[2] );

  /* Make the outputs */
  
  plhs[0] = mxCreateDoubleMatrix(1,coded,mxREAL);
  c = mxGetPr(plhs[0]);

  if ( nlhs==2 ){
    plhs[1] = mxCreateDoubleMatrix(1,N-M,mxREAL);
    info_map = mxGetPr(plhs[1]);

    //Write info map matlab readable
    for ( k = 0; k<N-M; k++ ) 
      info_map[k] = (double)(cols[k+M] +1); 
  }
    
  encode(u,c,info,blocks);

  mxFree(pchk_file);
  mxFree(gen_file);
  
  return;
}

void usage(void){ 
  mexPrintf("\nUsage:\n\n");
  mexPrintf("[ coded_bits [, message_bits_index] ]\n");
  mexPrintf("   = encode ( 'pchk_file', 'gen_file', info_bits )\n\n");
}
