clear
close all
rand('seed',4)  
randn('seed',4)

snr_dB = [ 1.9  2.2];
blocks = [20 50];

block_size = 100;
USE_GENERATOR = 1;
max_iterations = 200;

n = 1000;
m = 500;
pchk='ex-ldpc36-500.pchk';
gen='ex-ldpc36-500.gen';

r = 1-m/n;
snr = 10.^(snr_dB/10);
nvar = 1/2/r./snr;

uer = zeros(1,length(snr));
cer = zeros(1,length(snr));
wer = zeros(1,length(snr));

cum_blocks = cumsum([0,blocks]);
t0 = clock;

for l = 1:length(snr)
  fprintf('\nsnr_dB(%d) = %f\n',l,snr_dB(l));
  
  for p = 1:blocks(l)
    
    if (USE_GENERATOR==1)
    
      u = randint(1,(n-m)*block_size);
      [c,info] = ldpc_encode(pchk,gen,u);

      y = (-1).^c + sqrt(nvar(l))*randn(1,n*block_size); 
      a = 2*y/nvar(l);
  
      [b,blers,c_hat,u_hat] = ldpc_decode(pchk,max_iterations,a,info);
     
    else %all zero code word
      u = zeros(1,(n-m)*block_size);
      c = zeros(1,n*block_size);
      
      y = (-1).^c + sqrt(nvar(l))*randn(1,n*block_size); 
      a = 2*y/nvar(l);
         
      [b,blers,c_hat,u_hat] = ldpc_decode(pchk,max_iterations,a);      
     end

    wer(l) = (p-1)*wer(l)/p + sum(blers)/block_size/p;
    cer(l) = (p-1)*cer(l)/p + sum(ne(c,c_hat))/n/block_size/p;
    uer(l) = (p-1)*uer(l)/p + sum(ne(u,u_hat))/(n-m)/block_size/p;

    c_time = etime(clock,t0);
    a_time = c_time/(cum_blocks(l)+p)/60;
    r_time = (cum_blocks(length(snr)+1)-cum_blocks(l)-p)*a_time;
    
    fprintf('%5d: uer = %1.2e, cer = %1.2e, wer = %1.2e, (time=%fm)\n',p*block_size-1,uer(l),cer(l),wer(l),r_time);
  end

  fprintf('\nSNR = %1.2f dB: UER = %1.3e, CER = %1.3e, WER = %1.3e\n\n',snr_dB(l),uer(l),cer(l),wer(l));
end

SNR_DB = snr_dB
INFO_BIT_ERROR_RATE = uer 
CODE_BIT_ERROR_RATE = cer
WORD_ERROR_RATE = wer