/* Copyright 2007 Noah B. Jacobsen.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/************************************************************************/
/*        *******Includes code from:*********                           */
/*        Free software: Progressive edge-growth (PEG) algorithm        */
/*        Created by Xiaoyu Hu                                          */
/*                   Evangelos Eletheriou                               */
/*                   Dieter Arnold                                      */
/*        IBM Research, Zurich Research Lab., Switzerland               */
/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <math.h>
#include "BigTriang.h"
#include "NodesInGraph.h"
#include "Random.h"

BigTriang::BigTriang(void) {;}

BigTriang::BigTriang(int M, int N, int *symbolDegSequence, char *filename, Random *rand, int &numDeg, int *deg, double *degFrac){
  int i, j, k, m, index, localDepth=100, NumNodes=11*(N-M), *degSeq;
  
  int sglConcent=1;  //default to non-strictly concentrated parity-check distribution
  int tgtGirth=100000; // default to greedy PEG version 

  EXPAND_DEPTH=(tgtGirth-4)/2; 
  if(EXPAND_DEPTH<0) EXPAND_DEPTH=0;

  EXPAND_DEPTH=10;
  //sglConcent=numDeg<=0;
  sglConcent=1;

  (*this).M=M;
  (*this).N=N;
  (*this).filename=filename;
  H=NULL;

  localGirth=new int[NumNodes];
  nodesInGraph=new NodesInGraph [NumNodes];  

  for(i=0;i<N;i++)
    nodesInGraph[i].setNumOfConnectionSymbolBit(symbolDegSequence[i]);
 
  for(i=N;i<NumNodes;i++)
    nodesInGraph[i].setNumOfConnectionSymbolBit(0);

  double dtmp=0.0;
  degSeq = new int[M];

  for(i=0;i<numDeg;i++) dtmp+=degFrac[i];
  if(fabs(dtmp-1.0)>EPS) {
    cout.setf(ios::fixed, ios::floatfield);
    cout <<"\nInvalid degree distribution (node perspective): sum != 1.0 but "<<setprecision(10)<<dtmp<<endl; exit(-1); 
  } 

  for(i=1;i<numDeg;i++) degFrac[i]+=degFrac[i-1];

  for(i=0;i<M;i++) {
    dtmp=(double)i/M;
    for(j=numDeg-1;j>=0;j--) {
      if(dtmp>degFrac[j]) break;
    }
    if(dtmp<degFrac[0]) degSeq[i]=deg[0];
    else degSeq[i]=deg[j+1];
  }

  j=0; for(i=0;i<N;i++) j+=symbolDegSequence[i];
  k=0; for(i=0;i<M;i++) k+=degSeq[i];
  if(abs(k-j)>1000) {cout<<"\nBigTriang::BigTriang ERROR parity and symbol edge counts differ by 1000"<<endl;exit(-1);}

  deg[numDeg-1]=10000;
  for(i=0;i<M;i++){
    if(sglConcent==0) nodesInGraph[i].initConnectionParityBit(degSeq[i]);
    else nodesInGraph[i].initConnectionParityBit(); 
  }

  for(i=M;i<NumNodes-(N-M);i++)
    nodesInGraph[i].initConnectionParityBit(); 

  delete [] degSeq;
      
  for(k=N-1;k>=0;k--){
    if(0)
      //if(k>=N-M)
      nodesInGraph[k].connectionSymbolBit[0]=k-N+M;//diagonal parity bits

    else{
      m=1000000;index=-1;
      for(i=0;i<M;i++){
	if(nodesInGraph[i].numOfConnectionParityBit<m && nodesInGraph[i].numOfConnectionParityBit<nodesInGraph[i].maxDegParity) {
	  m=nodesInGraph[i].numOfConnectionParityBit;
	  index=i;
	}
      }
      nodesInGraph[k].connectionSymbolBit[0]=index;//least connections of parity bits
    }

    int iter=0; 

  ITER:
    localGirth[k]=100;
    for(m=1;m<nodesInGraph[k].numOfConnectionSymbolBit;m++){
 
      if(0)
	//if(k>=N-M)
	nodesInGraph[k].connectionSymbolBit[m]=selectConstrainedParityConnect(k, m, localDepth, M-N+k, rand, 1);
      else
	nodesInGraph[k].connectionSymbolBit[m]=selectConstrainedParityConnect(k, m, localDepth, 0, rand, sglConcent);

      localGirth[k]=(localGirth[k]>localDepth)?localDepth:localGirth[k];      
      if(k>0 && localGirth[k]<localGirth[k-1] && iter<20) {iter++; goto ITER;}
      if(localGirth[k]==0 && iter<30) {iter++; goto ITER;}
    }

    if((k+1)%100==0) {
      cout<<"k="<<k<<"  ";
      for(m=0;m<nodesInGraph[k].numOfConnectionSymbolBit;m++)
	cout<<nodesInGraph[k].connectionSymbolBit[m]<<" ";
      cout<<"LocalGirth="<<2*localGirth[k]+4;
      cout<<endl;
    }    
    
    updateConnection(k);
  }

  cout<<"Showing the row weight distribution..."<<endl;
  for(i=0;i<M;i++)
    cout<<nodesInGraph[i].numOfConnectionParityBit<<" ";
  cout<<endl;

  ofstream cycleFile;
  cycleFile.open("leftHandGirth.log", ios::out);
  localDepth=100;
  for(k=0;k<N;k++) {
    if(localGirth[k]<localDepth) localDepth=localGirth[k];
    if(localDepth==100) cycleFile<<"inf ";
    else cycleFile<<2*localDepth+4<<" ";
  }
  cycleFile<<endl;
  cycleFile.close();

  cout<<"*************************************************************"<<endl;
  cout<<"       The global girth of the PEG Tanner graph :="<< 2*localDepth+4<<endl;
  cout<<"*************************************************************"<<endl;

  //loadH();
  return;
}


BigTriang::~BigTriang(void) {
  if(H!=NULL) {
    for(int i=0;i<M;i++)
      delete [] H[i];
    delete [] H;
    H=NULL;
  }  
  delete [] localGirth;
  delete [] nodesInGraph;
  nodesInGraph=NULL;
}


int BigTriang::selectConstrainedParityConnect(int kthSymbol, int mthConnection, int & cycle, int startNewParity, Random *rand, int sglConcent) {
  int i, j, k, index, mincycles, numCur, numNext, cpNumCur;
  int *tmp,*indexSet;
  int *current;//take note of the covering parity bits

  mincycles=0;
  tmp=new int[M];
  indexSet=new int[25*M];
  
  numCur=mthConnection;
  current=new int[mthConnection];
  for(i=0;i<mthConnection;i++) current[i]=nodesInGraph[kthSymbol].connectionSymbolBit[i];

LOOP:
  mincycles++;
  for(i=0;i<M;i++) tmp[i]=0;
  //maintain 
  for(i=0;i<mthConnection;i++) tmp[nodesInGraph[kthSymbol].connectionSymbolBit[i]]=1;
  for(i=0;i<numCur;i++){
    for(j=0;j<nodesInGraph[current[i]].numOfConnectionParityBit;j++){
      for(k=0;k<nodesInGraph[nodesInGraph[current[i]].connectionParityBit[j]].numOfConnectionSymbolBit;k++){
        tmp[nodesInGraph[nodesInGraph[current[i]].connectionParityBit[j]].connectionSymbolBit[k]]=1;
      }
    }
  }

  index=0; cpNumCur=0;
  for(i=0;i<M;i++) {
    if(tmp[i]==1) cpNumCur++;
    if(i>=startNewParity && (tmp[i]==1 || nodesInGraph[i].numOfConnectionParityBit>=nodesInGraph[i].maxDegParity)) 
      index++;
  }

  if(index==M-startNewParity || mincycles>EXPAND_DEPTH){//covering all free parity nodes or meet the upper bound on cycles
    cycle=mincycles-1;
    for(i=0;i<M;i++) tmp[i]=0;
    for(i=0;i<numCur;i++) tmp[current[i]]=1;

    if(sglConcent==0){

      index=0;
      for(i=startNewParity;i<M;i++)
	if(tmp[i]==0 && nodesInGraph[i].numOfConnectionParityBit < nodesInGraph[i].maxDegParity)
	  for(j=nodesInGraph[i].numOfConnectionParityBit;j<nodesInGraph[i].maxDegParity;j++){
	    indexSet[index]=i;
	    
	    index++;
	  }
      
      j=(*rand).uniform(0,index);
      i=indexSet[j];
    }
  
    else{    
      index=0;
      for(i=startNewParity;i<M;i++) if(tmp[i]==1) index++;
      if(index>=M-startNewParity) {cout<<"Error in the case of index >= M-startNewParity (BigTriang::selectConstrainedParityConnect)"<<endl;exit(-1);}
      //additional handlement to select one having least connections
      j=10000000; 
      for(i=startNewParity;i<M;i++){
	if(tmp[i]==0 && nodesInGraph[i].numOfConnectionParityBit<j && nodesInGraph[i].numOfConnectionParityBit<nodesInGraph[i].maxDegParity)
	  j=nodesInGraph[i].numOfConnectionParityBit;
      }
      for(i=startNewParity;i<M;i++){
	if(tmp[i]==0){
	  if(nodesInGraph[i].numOfConnectionParityBit!=j || nodesInGraph[i].numOfConnectionParityBit>=nodesInGraph[i].maxDegParity){
	    tmp[i]=1;
	  }
	}
      }
      
      index=0;
      for(i=startNewParity;i<M;i++) if(tmp[i]==1) index++;
      
      j=(*rand).uniform(0, M-startNewParity-index)+1;
      index=0;
      for(i=startNewParity;i<M;i++){
	if(tmp[i]==0) index++;
	if(index==j) break;
      }
    }
     
    delete [] tmp; tmp=NULL;
    delete [] current; current=NULL;
    delete [] indexSet; indexSet=NULL;

    return(i);
  }

  else if(cpNumCur==numCur) {//can not expand any more
    if(sglConcent==0){

      index=0;
      for(i=startNewParity;i<M;i++)
	if(tmp[i]==0 && nodesInGraph[i].numOfConnectionParityBit < nodesInGraph[i].maxDegParity)
	  for(j=nodesInGraph[i].numOfConnectionParityBit;j<nodesInGraph[i].maxDegParity;j++){
	    indexSet[index]=i;

	    index++;
	  }

      j=(*rand).uniform(0,index);
      i=indexSet[j];
    }
  
    else{
      //additional handlement to select one having least connections
      j=10000000; //dummy number
      for(i=startNewParity;i<M;i++){
	if(tmp[i]==0 && nodesInGraph[i].numOfConnectionParityBit<j && nodesInGraph[i].numOfConnectionParityBit<nodesInGraph[i].maxDegParity)
	  j=nodesInGraph[i].numOfConnectionParityBit;
      }
      for(i=startNewParity;i<M;i++){
	if(tmp[i]==0){
	  if(nodesInGraph[i].numOfConnectionParityBit!=j || nodesInGraph[i].numOfConnectionParityBit>=nodesInGraph[i].maxDegParity){
	    tmp[i]=1;
	  }
	}
      }
      index=0;
      for(i=startNewParity;i<M;i++) if(tmp[i]==1) index++;
      //----------------------------------------------------------------
      j=(*rand).uniform(0, M-startNewParity-index)+1; //randomly selected
      index=0;
      for(i=startNewParity;i<M;i++){
	if(tmp[i]==0) index++;
	if(index==j) break;
      }
    }
    
    delete [] tmp; tmp=NULL;
    delete [] current; current=NULL;
    delete [] indexSet; indexSet=NULL;

    return(i);
  }
  
  else if(cpNumCur>numCur && index!=M-startNewParity){
    delete [] current;
    current=NULL;
    numCur=cpNumCur;
    current=new int[numCur];
    index=0;
    for(i=0;i<M;i++) {
      if(tmp[i]==1) {current[index]=i; index++;}
    }
    goto LOOP;
  }
  else {
    cout<<"Should not come to this point..."<<endl;
    cout<<"Error in BigTriang::selectParityConnect()"<<endl;
    return(-1);
  }
}


void BigTriang::updateConnection(int kthSymbol){
  int i, m;

  for(i=0;i<nodesInGraph[kthSymbol].numOfConnectionSymbolBit;i++){
    m=nodesInGraph[kthSymbol].connectionSymbolBit[i];//m [0, M) parity node

    nodesInGraph[m].connectionParityBit[nodesInGraph[m].numOfConnectionParityBit]=kthSymbol;
    nodesInGraph[m].numOfConnectionParityBit++; //increase by 1
  }
}


void BigTriang::loadH(void){
  int i, j;
  if(H==NULL) {
    H=new int*[M];
    for(i=0;i<M;i++) H[i]=new int[N];
  }

  for(i=0;i<M;i++){
    for(j=0;j<N;j++){
      H[i][j]=0;
    }
  }
  for(i=0;i<M;i++){
    for(j=0;j<nodesInGraph[i].numOfConnectionParityBit;j++){
      H[i][nodesInGraph[i].connectionParityBit[j]]=1;
    }
  }
}


void BigTriang::writeToFile(void){
  int i, j;
  FILE *f;

  f = fopen(filename,"wb");
  if (f==NULL)
    { fprintf(stderr,"Can't create parity check file: %s\n",filename);
    exit(1);
    }

  intio_write(f,('P'<<8)+0x80);
  if (ferror(f)){ 
  ERR:
    fprintf(stderr,"Error writing to parity check file: %s\n",filename);
    exit(1);
  }
  intio_write(f,M);
  if (ferror(f)) goto ERR;
  intio_write(f,N);
  if (ferror(f)) goto ERR;
    
  for(i=0;i<M;i++){
    intio_write (f, -(i+1));
    if (ferror(f)) goto ERR;

    for(j=0;j<nodesInGraph[i].numOfConnectionParityBit;j++){
      intio_write (f, nodesInGraph[i].connectionParityBit[j]+1);
      if (ferror(f)) goto ERR;
    }
  }
  
  intio_write(f,0);
  if (ferror(f)) goto ERR;
  if (fclose(f)!=0) goto ERR;
}


void BigTriang::intio_write
( FILE *f,  /* File to write to */
  int v     /* Value to write to file */
)
{ 
  unsigned char b;
  int i;

  for (i = 0; i<3; i++)
  { b = v&0xff;
    fwrite(&b,1,1,f);
    v >>= 8;
  }

  b = v>0 ? v : v+256;
  fwrite(&b,1,1,f);
}

