/* Copyright 2007 Noah B. Jacobsen.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <math.h>
#include "BigTriang.h"
#include "ExtTriang.h"
#include "NodesInGraph.h"
#include "Random.h"

ExtTriang::ExtTriang(void) {;}

ExtTriang::ExtTriang(int M, int N, int *symbolDegSequence, int numSplitParity, char *filename, BigTriang *baseGraph, Random *rand){
  int i, k, m, n, index, localDepth=100;
  int numNewParity, startPegParity, numPegParity, numAssignedParity;

  (*this).M=M;
  (*this).N=N;
  (*this).filename=filename;

  localGirth=new int[N];
  nodesInGraph=new NodesInGraph [N];

  for(n=0;n<N;n++) 
    nodesInGraph[n].setNumOfConnectionSymbolBit(symbolDegSequence[n]);

  for(m=0;m<M;m++) 
    nodesInGraph[m].initConnectionParityBit(); 

  mvBaseGraph(baseGraph, numNewParity);

  //*****If multiple levels of parity splitting are reqd******//
  numAssignedParity = (*baseGraph).M;
  while( numSplitParity > numAssignedParity ){
    //splitParityPeg(numNewParity, numAssignedParity, rand);
    splitParityRand(numNewParity, numAssignedParity, rand);
    
    numNewParity -= numAssignedParity;
    numSplitParity -= numAssignedParity;
    numAssignedParity += numAssignedParity;
  }
  //****************//

  //Splits the parities with weight metric
  //splitParityPeg(numNewParity, numSplitParity, rand);

  //Randomly splits the parities
  splitParityRand(numNewParity, numSplitParity, rand);

  numPegParity = numNewParity-numSplitParity;
  //cout << endl << "numPegParity = " << numPegParity << endl << endl;
  
 PEG:
  if(numPegParity>0){
    startPegParity = M-numPegParity;
    
    //Run PEG over remaining new parity symbols
    for(k=N-1;k>=0;k--){    
      int iter=0;

    ITER:
      localGirth[k]=100;
  
      for(m=nodesInGraph[k].numOfAssignedSymbolBit;m<nodesInGraph[k].numOfConnectionSymbolBit;m++){

	if(k >= N-startPegParity)
	  nodesInGraph[k].connectionSymbolBit[m]=selectConstrainedParityConnect(k, m, localDepth, k-N+M, rand);
	else
	  nodesInGraph[k].connectionSymbolBit[m]=selectConstrainedParityConnect(k, m, localDepth, startPegParity, rand);

	localGirth[k]=(localGirth[k]>localDepth)?localDepth:localGirth[k];
	if(k>0 && localGirth[k]<localGirth[k-1] && iter<20) {iter++; goto ITER;}
	if(localGirth[k]==0 && iter<30) {iter++; goto ITER;}
      }
      if((k+1)%100==0) {
	cout<<"k="<<k<<"  ";
	for(m=0;m<nodesInGraph[k].numOfConnectionSymbolBit;m++)
	  cout<<nodesInGraph[k].connectionSymbolBit[m]<<" ";
	cout<<"LocalGirth="<<2*localGirth[k]+4;
	cout<<endl;
      }    
    
      updateConnection(k);
    }
  }
  
  cout<<"Showing the row weight distribution..."<<endl;
  for(i=0;i<M;i++)
    cout<<nodesInGraph[i].numOfConnectionParityBit<<" ";
  cout<<endl;

  ofstream cycleFile;
  cycleFile.open("leftHandGirth.log", ios::out);
  localDepth=100;
  for(k=0;k<N;k++) {
    if(localGirth[k]<localDepth) localDepth=localGirth[k];
    if(localDepth==100) cycleFile<<"inf ";
    else cycleFile<<2*localDepth+4<<" ";
  }
  cycleFile<<endl;
  cycleFile.close();

  cout<<"*************************************************************"<<endl;
  cout<<"       The global girth of the PEG Tanner graph :="<< 2*localDepth+4<<endl;
  cout<<"*************************************************************"<<endl;

  mvExtGraph(baseGraph);
  return;
}

ExtTriang::~ExtTriang(void) {
  delete [] localGirth; localGirth=NULL;
  delete [] nodesInGraph; nodesInGraph=NULL;
}


int ExtTriang::selectConstrainedParityConnect(int kthSymbol, int mthConnection, int & cycle, int startPegParity, Random *rand) {
  int i, j, k, index, mincycles, numCur, numNext, cpNumCur;
  int *tmp;
  int *current;//take note of the covering parity bits

  mincycles=0;
  tmp=new int[M];

  numCur=mthConnection;
  current=new int[mthConnection];
  for(i=0;i<mthConnection;i++) current[i]=nodesInGraph[kthSymbol].connectionSymbolBit[i];

LOOP:
  mincycles++;
  for(i=0;i<M;i++) tmp[i]=0;
  //maintain 
  for(i=0;i<mthConnection;i++) tmp[nodesInGraph[kthSymbol].connectionSymbolBit[i]]=1;
  for(i=0;i<numCur;i++){
    for(j=0;j<nodesInGraph[current[i]].numOfConnectionParityBit;j++){
      for(k=0;k<nodesInGraph[nodesInGraph[current[i]].connectionParityBit[j]].numOfAssignedSymbolBit;k++){
        tmp[nodesInGraph[nodesInGraph[current[i]].connectionParityBit[j]].connectionSymbolBit[k]]=1;
      }
    }
  }

  index=0; cpNumCur=0;
  for(i=0;i<M;i++) {
    if(tmp[i]==1) cpNumCur++;
    if(i>=startPegParity && (tmp[i]==1 || nodesInGraph[i].numOfConnectionParityBit>=nodesInGraph[i].maxDegParity)) 
      index++;   
  }

  //covering all free parity nodes or meet the upper bound on cycles
  //if(index==M-startPegParity || mincycles>EXPAND_DEPTH){
  if(index==M-startPegParity){

    cycle=mincycles-1;
    for(i=0;i<M;i++) tmp[i]=0;
    for(i=0;i<numCur;i++) tmp[current[i]]=1;
    index=0;
    for(i=startPegParity;i<M;i++) if(tmp[i]==1) index++;
    if(index>=M-startPegParity) {cout<<"Error in the case of index >= M-startPegParity (ExtTriang::selectConstrainedParityConnect)"<<endl;exit(-1);}
    //additional handlement to select one having least connections
    j=10000000; 
    for(i=startPegParity;i<M;i++){
      if(tmp[i]==0 && nodesInGraph[i].numOfConnectionParityBit<j && nodesInGraph[i].numOfConnectionParityBit<nodesInGraph[i].maxDegParity)
	j=nodesInGraph[i].numOfConnectionParityBit;
    }
    for(i=startPegParity;i<M;i++){
      if(tmp[i]==0){
	if(nodesInGraph[i].numOfConnectionParityBit!=j || nodesInGraph[i].numOfConnectionParityBit>=nodesInGraph[i].maxDegParity){
	  tmp[i]=1;
	}
      }
    }
      
    index=0;
    for(i=startPegParity;i<M;i++) if(tmp[i]==1) index++;
   
    j=(*rand).uniform(0, M-startPegParity-index)+1;
    index=0;
    for(i=startPegParity;i<M;i++){
      if(tmp[i]==0) index++;
      if(index==j) break;
    }
    delete [] tmp; tmp=NULL;
    delete [] current; current=NULL;
    return(i);
  }

  else if(cpNumCur==numCur) {//can not expand any more
    //additional handlement to select one having least connections
    j=10000000; //dummy number
    for(i=startPegParity;i<M;i++){
      if(tmp[i]==0 && nodesInGraph[i].numOfConnectionParityBit<j && nodesInGraph[i].numOfConnectionParityBit<nodesInGraph[i].maxDegParity)
	j=nodesInGraph[i].numOfConnectionParityBit;
    }
    for(i=startPegParity;i<M;i++){
      if(tmp[i]==0){
	if(nodesInGraph[i].numOfConnectionParityBit!=j || nodesInGraph[i].numOfConnectionParityBit>=nodesInGraph[i].maxDegParity){
	  tmp[i]=1;
	}
      }
    }
    index=0;
    for(i=startPegParity;i<M;i++) if(tmp[i]==1) index++;
    //----------------------------------------------------------------
    j=(*rand).uniform(0, M-startPegParity-index)+1; //randomly selected
    index=0;
    for(i=startPegParity;i<M;i++){
      if(tmp[i]==0) index++;
      if(index==j) break;
    }
    delete [] tmp; tmp=NULL;
    delete [] current; current=NULL;
    return(i);
  }

  else if(cpNumCur>numCur && index!=M-startPegParity){
    delete [] current;
    current=NULL;
    numCur=cpNumCur;
    current=new int[numCur];
    index=0;
    for(i=0;i<M;i++) {
      if(tmp[i]==1) {current[index]=i; index++;}
    }
    goto LOOP;
  }
  else {
    cout<<"Should not come to this point..."<<endl;
    cout<<"Error in ExtTriang::selectParityConnect()"<<endl;
    return(-1);
  }
}

void ExtTriang::updateConnection(int kthSymbol){
  int i, m;

  for(i=nodesInGraph[kthSymbol].numOfAssignedSymbolBit;i<nodesInGraph[kthSymbol].numOfConnectionSymbolBit;i++){
    m=nodesInGraph[kthSymbol].connectionSymbolBit[i];//m [0, M) parity node

    nodesInGraph[m].connectionParityBit[nodesInGraph[m].numOfConnectionParityBit]=kthSymbol;
    nodesInGraph[m].numOfConnectionParityBit++;
  }
}


void ExtTriang::splitParityPeg(int numNewParity, int numSplitParity, Random *rand){
  int i, j, k, l, m, n;
  int *mv, *tmp, *current, *origParityConnection;
  int numOrig, numMove, numLeft, numCur;
  int newParityIndex, newSymbolIndex;
  int index, mvSymbol, *revSortParityIndex;
  int curParityBit, curSymbolBit;
  int curDepth, maxGirth, maxWeight, minWeight, minGirth;
  int *girth, *weight, *curWeight, wgt;
  int Moved, newNumCur, maxDegParity=60, MaxDepth=10, WeightThreshold=8, maxInitWgt;
  int symbolBit, parityBit, *connectionParityBit, numUnderWeight=0, totMoved=0;

  tmp=new int[N];
  curWeight=new int[N];
  connectionParityBit=new int[N];
  revSortParityIndex=new int[M-numNewParity];

  index=0;
  for(i=maxDegParity;i>=1;i--)
    for(m=0;m<M-numNewParity;m++)
    //for(m=M-numNewParity-1;m>=0;m--)
      if(nodesInGraph[m].numOfConnectionParityBit==i){
	revSortParityIndex[index]=m;
	index++;
      }

  if(index!=M-numNewParity){
    cout <<"Error in splitParity::index = "<<index<<", M-numNewParity = "<<M-numNewParity<<"\n"<<endl;exit(-1);}  

  newParityIndex=M-numNewParity;
  newSymbolIndex=N-numNewParity;

  for(m=0;m<numSplitParity;m++){
    curParityBit=revSortParityIndex[m];
    
    numOrig = nodesInGraph[curParityBit].numOfConnectionParityBit;    
    numMove = (int) floor((double) numOrig/2);//Will move less bits than keep

    //First insert new degree-two symbol node
    nodesInGraph[m+newParityIndex].numOfConnectionParityBit=1;
    nodesInGraph[m+newParityIndex].connectionParityBit[0]=m+newSymbolIndex;
    
    nodesInGraph[curParityBit].connectionParityBit[numOrig]=m+newSymbolIndex;
    nodesInGraph[curParityBit].numOfConnectionParityBit++;

    // New degree-two node
    nodesInGraph[m+newSymbolIndex].numOfAssignedSymbolBit=2;
    nodesInGraph[m+newSymbolIndex].connectionSymbolBit[0]=curParityBit;
    nodesInGraph[m+newSymbolIndex].connectionSymbolBit[1]=m+newParityIndex;
    nodesInGraph[m+newSymbolIndex].keepAsDegreeTwo=1;
    
    // Copy current parity
    origParityConnection=new int[numOrig];
    for(j=0;j<numOrig;j++)
      origParityConnection[j]=nodesInGraph[curParityBit].connectionParityBit[j];

    // Loop over selecting a symbol to move
    for( Moved=0;Moved<numMove;Moved++){
      numLeft=numOrig-Moved;
      
      maxWeight=0; minGirth=100000;
      maxGirth=0; minWeight=100000;
      weight=new int[numLeft];
      girth=new int[numLeft];
      
      //-------------------------Compute weight
      for(n=0;n<numLeft;n++){
	curSymbolBit=origParityConnection[n];
	splitSymbol(curSymbolBit, curParityBit, m+newParityIndex);
	
	for(i=0;i<N;i++){ tmp[i]=0; curWeight[i]=0;}

	numCur=numLeft-1;

	//cout << "numCur = " << numCur << ", " << nodesInGraph[curParityBit].numOfConnectionParityBit << endl;
	//exit(-1);

	maxInitWgt=0;
	for(i=0;i<numCur;i++){
	  symbolBit = nodesInGraph[curParityBit].connectionParityBit[i];
	  tmp[symbolBit]=1;
	  
	  curWeight[symbolBit]=nodesInGraph[symbolBit].numOfAssignedSymbolBit-2;
	  connectionParityBit[symbolBit]=curParityBit;

	  if( maxInitWgt< curWeight[symbolBit])
	    maxInitWgt = curWeight[symbolBit];
	}
	if( maxInitWgt< WeightThreshold) maxInitWgt = WeightThreshold;
	
	curDepth=0;
      LOOP:
	if(curDepth>MaxDepth){cout<<"\nError in ExtTriang::splitParityPeg curDepth out of bounds\n"<<endl;exit(-1);}

	curDepth++;
	current=new int[numCur];
	
	index=0;
	//Expand low weight nodes first
	for( wgt=-1; wgt<=maxInitWgt; wgt++)	
	  for(i=0;i<N;i++)
	    if( tmp[i]==1 && curWeight[i]==wgt){
	      current[index]=i;
	      index++;
	    }
      
	//cout << "maxInitWgt = " << maxInitWgt << endl;
	//cout << "index = " << index << endl;
	//cout << "numCur = " << numCur << endl;

	if(index!=numCur){cout<<"\nError in ExtTriang::splitParityPeg where index!=numCur\n"<<endl;exit(-1);}
	
	for(i=0;i<N;i++) tmp[i]=0;
	newNumCur=0;

	for(i=0;i<numCur;i++){
	  
	  for(j=0;j<nodesInGraph[current[i]].numOfAssignedSymbolBit;j++){
	    parityBit=nodesInGraph[current[i]].connectionSymbolBit[j];

	    if(parityBit==curParityBit && parityBit!=connectionParityBit[current[i]]){
	      girth[n]=curDepth;
	      weight[n]=curWeight[current[i]];
		  
	      goto NEXT;
	    }
	    
	    else if(parityBit!=connectionParityBit[current[i]])

	      for(k=0;k<nodesInGraph[parityBit].numOfConnectionParityBit;k++){
		symbolBit=nodesInGraph[parityBit].connectionParityBit[k];
		
		if(symbolBit!=current[i]){//min weight of connecting parity bits
		  wgt=curWeight[current[i]] + nodesInGraph[symbolBit].numOfAssignedSymbolBit-2;
		  
		  if(tmp[symbolBit]==0 && wgt<WeightThreshold){
		    tmp[symbolBit]=1;
		    newNumCur++;
		    
		    curWeight[symbolBit] = wgt;  
		    connectionParityBit[symbolBit]=parityBit;
		  }

		  else if(wgt<curWeight[symbolBit]){
		    curWeight[symbolBit] = wgt;
		    connectionParityBit[symbolBit]=parityBit;
		  }
		}
	      }
	  }
	}
	
	delete [] current; current=NULL;
	numCur=newNumCur;
	
	if(curDepth==MaxDepth){
	  girth[n] = MaxDepth;
	  weight[n] = WeightThreshold;
	  
	  goto NEXT;
	}
	
	goto LOOP;
	
      NEXT:
	unsplitSymbol(curSymbolBit, curParityBit, m+newParityIndex);

	maxGirth = (girth[n]>maxGirth) ? girth[n] : maxGirth;
	minGirth = (girth[n]>maxGirth) ? maxGirth : girth[n];
	maxWeight = (weight[n]>maxWeight) ? weight[n] : maxWeight;
	minWeight = (weight[n]>minWeight) ? minWeight : weight[n];
      }
      //-----------------------Completed metric calculation for the numLeft parity symbols
      //pick random maxWeight symbol
      if(maxWeight<WeightThreshold)
	numUnderWeight++;

      wgt=0;
      for(k=0;k<numLeft;k++)
	if(weight[k]==maxWeight)
	  wgt++;
      
      n = (*rand).uniform(0,wgt);

      wgt=0;
      for(k=0;k<numLeft;k++)
	if(weight[k]==maxWeight){
	  if(n==wgt){
	    mvSymbol=origParityConnection[k];
	    
	    goto MOVE;
	  }
	  else
	    wgt++;
	}

    MOVE:
      delete [] girth; girth=NULL;
      delete [] weight; weight=NULL;

      splitSymbol(mvSymbol, curParityBit, m+newParityIndex);

      index=0;
      for(j=0;j<numLeft;j++)
	if(origParityConnection[j]!=mvSymbol){
	  origParityConnection[index]=origParityConnection[j];
	  index++;
	}
      if(index!=numLeft-1){cout<<"\nError in ExtTriang::splitParityPeg where mvSymbol not found in origParityConnection\n"<<endl;exit(-1);}
    }
   
    delete [] origParityConnection; origParityConnection=NULL;
    totMoved+=numMove;
  }

  cout <<"\nSplitParityPeg::"<<filename<<" numUnderWeight = "<<numUnderWeight<<"/"<<totMoved<<"\n"<<endl;

  delete [] tmp; tmp=NULL;
  delete [] curWeight; curWeight=NULL;
  delete [] connectionParityBit; connectionParityBit=NULL;
  delete [] revSortParityIndex; revSortParityIndex=NULL;

  return;
}

void ExtTriang::splitSymbol(int mvSymbol, int origParity, int newParity){
  int k,index;
  int numOrig, numNew;

  // Keep de-selected edges
  numOrig=nodesInGraph[origParity].numOfConnectionParityBit;

  index=0;
  for(k=0;k<numOrig;k++)
    if(nodesInGraph[origParity].connectionParityBit[k]!=mvSymbol){
      nodesInGraph[origParity].connectionParityBit[index]=nodesInGraph[origParity].connectionParityBit[k];
      index++;
    }
  if(index!=numOrig-1){cout<<"\nError in ExtTriang::splitSymbol where index!=numOrig-1\n"<<endl;exit(-1);}
  nodesInGraph[origParity].numOfConnectionParityBit=numOrig-1;
      
  // Move selected edge
  numNew=nodesInGraph[newParity].numOfConnectionParityBit;

  nodesInGraph[newParity].connectionParityBit[numNew]=mvSymbol;
  nodesInGraph[newParity].numOfConnectionParityBit++;
      
  for(index=0;index<nodesInGraph[mvSymbol].numOfConnectionSymbolBit;index++)
    if(nodesInGraph[mvSymbol].connectionSymbolBit[index]==origParity){
      nodesInGraph[mvSymbol].connectionSymbolBit[index]=newParity;
      break;
    }
  if(index==nodesInGraph[mvSymbol].numOfConnectionSymbolBit){cout<<"\nError in ExtTriang::splitSymbol where mvSymbol orig. parity not found\n"<<endl;exit(-1);}
  
  return;
}

void ExtTriang::unsplitSymbol(int mvSymbol, int origParity, int newParity){
  int k,index;
  int numOrig, numNew;

  // Keep de-selected edges
  numOrig=nodesInGraph[newParity].numOfConnectionParityBit;

  index=0;
  for(k=0;k<numOrig;k++)
    if(nodesInGraph[newParity].connectionParityBit[k]!=mvSymbol){
      nodesInGraph[newParity].connectionParityBit[index]=nodesInGraph[newParity].connectionParityBit[k];
      index++;
    }
  if(index!=numOrig-1){cout<<"\nError in ExtTriang::unsplitSymbol where index!=numOrig-1\n"<<endl;exit(-1);}
  nodesInGraph[newParity].numOfConnectionParityBit=numOrig-1;
      
  // Move selected edge
  numNew=nodesInGraph[origParity].numOfConnectionParityBit;

  nodesInGraph[origParity].connectionParityBit[numNew]=mvSymbol;
  nodesInGraph[origParity].numOfConnectionParityBit++;
      
  for(index=0;index<nodesInGraph[mvSymbol].numOfConnectionSymbolBit;index++)
    if(nodesInGraph[mvSymbol].connectionSymbolBit[index]==newParity){
      nodesInGraph[mvSymbol].connectionSymbolBit[index]=origParity;
      break;
    }
  if(index==nodesInGraph[mvSymbol].numOfConnectionSymbolBit){cout<<"\nError in ExtTriang::unsplitSymbol where mvSymbol orig. parity not found\n"<<endl;exit(-1);}
  
  return;
}


void ExtTriang::splitParityRand(int numNewParity, int numSplitParity, Random *rand){
  int i, j, m, n, maxDegParity=100;
  int rn, tmp, *mv;
  int numOrig, numMove, numKeep, *origParityConnection;
  int newParityIndex, newSymbolIndex;
  int curParityBit, index, mvSymbol, *revSortParityIndex;

  revSortParityIndex=new int[M-numNewParity];

  index=0;
  for(i=maxDegParity;i>=1;i--)
    for(m=0;m<M-numNewParity;m++)
      if(nodesInGraph[m].numOfConnectionParityBit==i){
	revSortParityIndex[index]=m;
	index++;
      }

  if(index!=M-numNewParity){
    cout <<"Error in splitParity::index = "<<index<<", M-numNewParity = "<<M-numNewParity<<"\n"<<endl;}  

  newParityIndex=M-numNewParity;
  newSymbolIndex=N-numNewParity;

  mv=new int[maxDegParity];
  for(m=0;m<numSplitParity;m++){
    curParityBit=revSortParityIndex[m];

    numOrig = nodesInGraph[curParityBit].numOfConnectionParityBit;
    numMove = (int) ceil((double) numOrig/2);
    numKeep = numOrig-numMove;

    for(j=0;j<maxDegParity;j++) mv[j]=j;

    for(i=0;i<numKeep;i++){
      rn = i + (*rand).uniform(0,numOrig-i);

      tmp = mv[i];
      mv[i] = mv[rn];
      mv[rn] = tmp;
    }
  
    // Copy original row
    origParityConnection=new int[numOrig];
    for(j=0;j<numOrig;j++)
      origParityConnection[j]=nodesInGraph[curParityBit].connectionParityBit[j];

    // Keep selected edges
    nodesInGraph[curParityBit].numOfConnectionParityBit=numKeep+1;

    for(i=0;i<numKeep;i++)
      nodesInGraph[curParityBit].connectionParityBit[i]=origParityConnection[mv[i]];
    nodesInGraph[curParityBit].connectionParityBit[numKeep]=m+newSymbolIndex;

    nodesInGraph[m+newSymbolIndex].numOfAssignedSymbolBit=2;
    nodesInGraph[m+newSymbolIndex].connectionSymbolBit[0]=curParityBit;
    nodesInGraph[m+newSymbolIndex].connectionSymbolBit[1]=m+newParityIndex;
    nodesInGraph[m+newSymbolIndex].keepAsDegreeTwo=1;

    // Move other edges
    nodesInGraph[m+newParityIndex].numOfConnectionParityBit=numMove+1;
    nodesInGraph[m+newParityIndex].connectionParityBit[numMove]=m+newSymbolIndex;
 
    for(j=0;j<numMove;j++){
      mvSymbol = origParityConnection[mv[numKeep+j]];
      nodesInGraph[m+newParityIndex].connectionParityBit[j]=mvSymbol;

      for(index=0;index<nodesInGraph[mvSymbol].numOfConnectionSymbolBit;index++)
	if(nodesInGraph[mvSymbol].connectionSymbolBit[index]==curParityBit){
	  nodesInGraph[mvSymbol].connectionSymbolBit[index]=m+newParityIndex;
	  break;
	}
      if(index==nodesInGraph[mvSymbol].numOfConnectionSymbolBit){cout<<"\n Can't find original parity index\n"<<endl;exit(-1);}
    }      

    delete [] origParityConnection; origParityConnection=NULL;
  }
  delete [] revSortParityIndex; revSortParityIndex=NULL;
  delete [] mv; mv=NULL;

  return;
}


void ExtTriang::mvBaseGraph(BigTriang *baseGraph, int &M1){
  int i, k, m, n;
  int M0;

  M0 = (*baseGraph).M;
  M1 = M-M0;

  for (n=0; n<(*baseGraph).N; n++){
    nodesInGraph[n].numOfAssignedSymbolBit = (*baseGraph).nodesInGraph[n].numOfConnectionSymbolBit;
    nodesInGraph[n].keepAsDegreeTwo = (*baseGraph).nodesInGraph[n].keepAsDegreeTwo;

    if(nodesInGraph[n].numOfAssignedSymbolBit>nodesInGraph[n].numOfConnectionSymbolBit){cout<<"\n numOfAssignedSymbolBit>numOfConnectionSymbolBit\n"<<endl;exit(-1);} 

    for (m=0; m<nodesInGraph[n].numOfAssignedSymbolBit; m++)
      nodesInGraph[n].connectionSymbolBit[m] = (*baseGraph).nodesInGraph[n].connectionSymbolBit[m];      
  }

  for (m=0; m<M0; m++){
    nodesInGraph[m].numOfConnectionParityBit = (*baseGraph).nodesInGraph[m].numOfConnectionParityBit;
 
    for (n=0; n<nodesInGraph[m].numOfConnectionParityBit; n++)
      nodesInGraph[m].connectionParityBit[n] = (*baseGraph).nodesInGraph[m].connectionParityBit[n];    
  }

  return;
}


void ExtTriang::mvExtGraph(BigTriang *bigTriang){
  int n,m,Deg;
  
  (*bigTriang).N = N;
  (*bigTriang).M = M;
  (*bigTriang).filename=filename;

  for(n=0;n<N;n++){
    Deg = nodesInGraph[n].numOfConnectionSymbolBit;

    (*bigTriang).nodesInGraph[n].numOfConnectionSymbolBit = Deg;
    (*bigTriang).nodesInGraph[n].keepAsDegreeTwo = nodesInGraph[n].keepAsDegreeTwo;
    
    for(m=0;m<Deg;m++)
      (*bigTriang).nodesInGraph[n].connectionSymbolBit[m]=nodesInGraph[n].connectionSymbolBit[m];
  }

  for(m=0;m<M;m++){
    Deg = nodesInGraph[m].numOfConnectionParityBit;
    
    (*bigTriang).nodesInGraph[m].numOfConnectionParityBit = Deg;
     
    for(n=0;n<Deg;n++)
      (*bigTriang).nodesInGraph[m].connectionParityBit[n]=nodesInGraph[m].connectionParityBit[n];
  }

  for(n=0;n<N;n++)
    (*bigTriang).localGirth[n]=localGirth[n];

  return;
}
